/*-------------------------------------------------------------------------
 *
 * deparse.c
 *		  query deparser for PostgreSQL
 *
 * Copyright (c) 2012, PostgreSQL Global Development Group
 *
 * IDENTIFICATION
 *		  contrib/pgsql_fdw/deparse.c
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"

#include "access/transam.h"
#include "foreign/foreign.h"
#include "lib/stringinfo.h"
#include "nodes/nodeFuncs.h"
#include "nodes/nodes.h"
#include "nodes/makefuncs.h"
#include "optimizer/clauses.h"
#include "optimizer/var.h"
#include "parser/parsetree.h"
#include "utils/builtins.h"
#include "utils/lsyscache.h"

#include "pgsql_fdw.h"

/*
 * Context for walk-through the expression tree.
 */
typedef struct foreign_executable_cxt
{
	PlannerInfo	   *root;
	RelOptInfo	   *foreignrel;
} foreign_executable_cxt;

/*
 * Deparse query representation into SQL statement which suits for remote
 * PostgreSQL server.
 */
char *
deparseSql(Oid relid, PlannerInfo *root, RelOptInfo *baserel)
{
	StringInfoData	foreign_relname;
	StringInfoData	sql;			/* builder for SQL statement */
	bool		first;
	AttrNumber	attr;
	List	   *attr_used = NIL;	/* List of AttNumber used in the query */
	const char *nspname = NULL;		/* plain namespace name */
	const char *relname = NULL;		/* plain relation name */
	const char *q_nspname;			/* quoted namespace name */
	const char *q_relname;			/* quoted relation name */
	int			i;
	List	   *rtable = NIL;
	List	   *context = NIL;

	initStringInfo(&sql);
	initStringInfo(&foreign_relname);

	/*
	 * First of all, determine which column should be retrieved for this scan.
	 *
	 * We do this before deparsing SELECT clause because attributes which are
	 * not used in neither reltargetlist nor baserel->baserestrictinfo, quals
	 * evaluated on local, can be replaced with literal "NULL" in the SELECT
	 * clause to reduce overhead of tuple handling tuple and data transfer.
	 */
	if (baserel->baserestrictinfo != NIL)
	{
		ListCell   *lc;

		foreach (lc, baserel->baserestrictinfo)
		{
			RestrictInfo   *ri = (RestrictInfo *) lfirst(lc);
			List		   *attrs;

			/*
			 * We need to know which attributes are used in qual evaluated
			 * on the local server, because they should be listed in the
			 * SELECT clause of remote query.  We can ignore attributes
			 * which are referenced only in ORDER BY/GROUP BY clause because
			 * such attributes has already been kept in reltargetlist.
			 */
			attrs = pull_var_clause((Node *) ri->clause,
									PVC_RECURSE_AGGREGATES,
									PVC_RECURSE_PLACEHOLDERS);
			attr_used = list_union(attr_used, attrs);
		}
	}

	/*
	 * Determine foreign relation's qualified name.  This is necessary for
	 * FROM clause and SELECT clause.
	 */
	nspname = GetFdwOptionValue(InvalidOid, InvalidOid, relid,
								InvalidAttrNumber, "nspname");
	if (nspname == NULL)
		nspname = get_namespace_name(get_rel_namespace(relid));
	q_nspname = quote_identifier(nspname);

	relname = GetFdwOptionValue(InvalidOid, InvalidOid, relid,
								InvalidAttrNumber, "relname");
	if (relname == NULL)
		relname = get_rel_name(relid);
	q_relname = quote_identifier(relname);

	appendStringInfo(&foreign_relname, "%s.%s", q_nspname, q_relname);

	/*
	 * We need to replace aliasname and colnames of the target relation so that
	 * constructed remote query is valid.
	 * 
	 * Note that we skip first empty element of simple_rel_array.  See also
	 * comments of simple_rel_array and simple_rte_array for the rationale.
	 */
	for (i = 1; i < root->simple_rel_array_size; i++)
	{
		RangeTblEntry  *rte = copyObject(root->simple_rte_array[i]);
		List		   *newcolnames = NIL;

		if (i == baserel->relid)
		{
			/*
			 * Create new list of column names which is used to deparse remote
			 * query from specified names or local column names.  This list is
			 * used by deparse_expression.
			 */
			for (attr = 1; attr <= baserel->max_attr; attr++)
			{
				char	   *colname = NULL;

				/*
				 * Get colname option for each undropped attribute.  If colname
				 * option is not supplied, use local column name in remote
				 * query.
				 */
				if (get_rte_attribute_is_dropped(rte, attr))
					colname = "";
				else
				{
					colname = GetFdwOptionValue(InvalidOid, InvalidOid, relid,
												attr, "colname");
					if (colname == NULL)
						colname = strVal(list_nth(rte->eref->colnames,
												  attr - 1));
				}
				newcolnames = lappend(newcolnames, makeString(colname));
			}
			rte->alias = makeAlias(relname, newcolnames);
		}
		rtable = lappend(rtable, rte);
	}
	context = deparse_context_for_rtelist(rtable);

	/*
	 * deparse SELECT clause
	 *
	 * List attributes which are in either target list or local restriction.
	 * Unused attributes are replaced with a literal "NULL" for optimization.
	 *
	 * Note that nothing is added for dropped columns, though tuple constructor
	 * function requires entries for dropped columns.  Such entries must be
	 * initialized with NULL before calling tuple constructor.
	 */
	appendStringInfo(&sql, "SELECT ");
	attr_used = list_union(attr_used, baserel->reltargetlist);
	first = true;
	for (attr = 1; attr <= baserel->max_attr; attr++)
	{
		RangeTblEntry *rte = root->simple_rte_array[baserel->relid];
		Var		   *var = NULL;
		ListCell   *lc;

		/* Ignore dropped attributes. */
		if (get_rte_attribute_is_dropped(rte, attr))
			continue;

		if (!first)
			appendStringInfo(&sql, ", ");
		first = false;

		/*
		 * We use linear search here, but it wouldn't be problem since
		 * attr_used seems to not become so large.
		 */
		foreach (lc, attr_used)
		{
			var = lfirst(lc);
			if (var->varattno == attr)
				break;
			var = NULL;
		}
		if (var != NULL)
			appendStringInfo(&sql, "%s",
				deparse_expression((Node *) var, context, false, false));
		else
			appendStringInfo(&sql, "NULL");
	}
	appendStringInfoChar(&sql, ' ');

	/*
	 * deparse FROM clause, including alias if any
	 */
	appendStringInfo(&sql, "FROM %s", foreign_relname.data);

	elog(DEBUG3, "Remote SQL: %s", sql.data);
	return sql.data;
}

