/*-------------------------------------------------------------------------
 *
 * pgsql_fdw.c
 *		  foreign-data wrapper for remote PostgreSQL servers.
 *
 * Copyright (c) 2012, PostgreSQL Global Development Group
 *
 * IDENTIFICATION
 *		  contrib/pgsql_fdw/pgsql_fdw.c
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"
#include "fmgr.h"

#include "access/transam.h"
#include "catalog/pg_foreign_server.h"
#include "catalog/pg_foreign_table.h"
#include "commands/explain.h"
#include "commands/vacuum.h"
#include "foreign/fdwapi.h"
#include "funcapi.h"
#include "miscadmin.h"
#include "nodes/nodeFuncs.h"
#include "optimizer/cost.h"
#include "optimizer/pathnode.h"
#include "parser/parse_relation.h"
#include "parser/parse_type.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "utils/lsyscache.h"
#include "utils/memutils.h"
#include "utils/rel.h"
#include "utils/syscache.h"

#include "pgsql_fdw.h"
#include "connection.h"

PG_MODULE_MAGIC;

/*
 * Default fetch count for cursor.  This can be overridden by fetch_count FDW
 * option.
 */
#define DEFAULT_FETCH_COUNT			10000

/*
 * Cost to establish a connection.
 * XXX: should be configurable per server?
 */
#define CONNECTION_COSTS			100.0

/*
 * Cost to transfer 1 byte from remote server.
 * XXX: should be configurable per server?
 */
#define TRANSFER_COSTS_PER_BYTE		0.001

/*
 * Cursors which are used together in a local query require different name, so
 * we use simple incremental name for that purpose.  We don't care wrap around
 * of cursor_id because it's hard to imagine that 2^32 cursors are used in a
 * query.
 */
#define	CURSOR_NAME_FORMAT		"pgsql_fdw_cursor_%u"
static uint32 cursor_id = 0;

/*
 * Utility macros for quick access to the first row in PGresult
 */
#define PGRES_VAL0(col)		PQgetvalue(res, 0, (col))
#define PGRES_NULL0(col)	PQgetisnull(res, 0, (col))

/*
 * Index of FDW-private items stored in FdwPlan.
 */
enum FdwPrivateIndex {
	FdwPrivateSelectSql,
	FdwPrivateDeclareSql,
	FdwPrivateFetchSql,
	FdwPrivateResetSql,
	FdwPrivateCloseSql,

	/* # of elements stored in the list fdw_private */
	FdwPrivateNum,
};

/*
 * Describes an execution state of a foreign scan against a foreign table
 * using pgsql_fdw. 
 */
typedef struct PgsqlFdwExecutionState
{
	MemoryContext scan_cxt;		/* context for per-scan lifespan data */

	FdwPlan	   *fdwplan;		/* FDW-specific planning information */
	PGconn	   *conn;			/* connection for the scan */

	Oid		   *param_types;	/* type array of external parameter */
	const char **param_values;	/* value array of external parameter */

	int			attnum;			/* # of non-dropped attribute */
	char	  **col_values;		/* column value buffer */
	AttInMetadata *attinmeta;	/* attribute metadata */

	Tuplestorestate *tuples;	/* result of the scan */
} PgsqlFdwExecutionState;

/*
 * SQL functions
 */
extern Datum pgsql_fdw_handler(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(pgsql_fdw_handler);

/*
 * FDW callback routines
 */
static FdwPlan *pgsqlPlanForeignScan(Oid foreigntableid,
									 PlannerInfo *root,
									 RelOptInfo *baserel);
static void pgsqlExplainForeignScan(ForeignScanState *node, ExplainState *es);
static void pgsqlBeginForeignScan(ForeignScanState *node, int eflags);
static TupleTableSlot *pgsqlIterateForeignScan(ForeignScanState *node);
static void pgsqlReScanForeignScan(ForeignScanState *node);
static void pgsqlEndForeignScan(ForeignScanState *node);
static void pgsqlAnalyzeForeignTable(Relation relation,
									 VacuumStmt *vacstmt,
									 int elevel);

/*
 * Helper functions
 */
static void estimate_costs(PlannerInfo *root,
						   RelOptInfo *baserel,
						   const char *sql,
						   Oid serverid,
						   Cost *startup_cost,
						   Cost *total_cost);
static void execute_query(ForeignScanState *node);
static PGresult *fetch_result(ForeignScanState *node);
static void store_result(ForeignScanState *node, PGresult *res);
static bool contain_ext_param(Node *clause);
static bool contain_ext_param_walker(Node *node, void *context);
static void store_remote_stats(PGresult *res, VacAttrStats *stats);

/* Exported functions, but not written in pgsql_fdw.h. */
void _PG_init(void);
void _PG_fini(void);

/*
 * Module-specific initialization.
 */
void
_PG_init(void)
{
}

/*
 * Module-specific clean up.
 */
void
_PG_fini(void)
{
}

/*
 * The content of FdwRoutine of pgsql_fdw is never changed, so we use one
 * static object for all scan.
 */
static FdwRoutine fdwroutine = {

	/* Node type */
	T_FdwRoutine,

	/* Required handler functions. */
	pgsqlPlanForeignScan,
	pgsqlExplainForeignScan,
	pgsqlBeginForeignScan,
	pgsqlIterateForeignScan,
	pgsqlReScanForeignScan,
	pgsqlEndForeignScan,

	/* Optional handler functions. */
	pgsqlAnalyzeForeignTable,
};

/*
 * Foreign-data wrapper handler function: return a struct with pointers
 * to my callback routines.
 */
Datum
pgsql_fdw_handler(PG_FUNCTION_ARGS)
{
	PG_RETURN_POINTER(&fdwroutine);
}

/*
 * pgsqlPlanForeignScan
 *		Create a FdwPlan for a scan on the foreign table
 */
static FdwPlan *
pgsqlPlanForeignScan(Oid foreigntableid,
					PlannerInfo *root,
					RelOptInfo *baserel)
{
	char			name[128];	/* must be larger than format + 10 */
	StringInfoData	cursor;
	const char	   *fetch_count_str;
	int				fetch_count = DEFAULT_FETCH_COUNT;
	char		   *sql;
	FdwPlan		   *fdwplan;
	List		   *fdw_private = NIL;
	ForeignTable   *table;
	ForeignServer  *server;

	/* Construct FdwPlan with cost estimates */
	fdwplan = makeNode(FdwPlan);
	sql = deparseSql(foreigntableid, root, baserel);
	table = GetForeignTable(foreigntableid);
	server = GetForeignServer(table->serverid);
	estimate_costs(root, baserel, sql, server->serverid,
				   &fdwplan->startup_cost, &fdwplan->total_cost);

	/*
	 * Store plain SELECT statement in private area of FdwPlan.  This will be
	 * used for executing remote query and explaining scan.
	 */
	fdw_private = list_make1(makeString(sql));

	/* Use specified fetch_count instead of default value, if any. */
	fetch_count_str = GetFdwOptionValue(InvalidOid, InvalidOid, foreigntableid,
										InvalidAttrNumber, "fetch_count");
	if (fetch_count_str != NULL)
		fetch_count = strtol(fetch_count_str, NULL, 10);
	elog(DEBUG1, "relid=%u fetch_count=%d", foreigntableid, fetch_count);

	/*
	 * We store some more information in FdwPlan to pass them beyond the
	 * boundary between planner and executor.  Finally FdwPlan using cursor
	 * would hold items below:
	 *
	 * 1) plain SELECT statement (already added above)
	 * 2) SQL statement used to declare cursor
	 * 3) SQL statement used to fetch rows from cursor
	 * 4) SQL statement used to reset cursor 
	 * 5) SQL statement used to close cursor
	 *
	 * These items are indexed with the enum FdwPrivateIndex, so an item
	 * can be accessed directly via list_nth().  For example of FETCH
	 * statement:
	 *      list_nth(fdw_private, FdwPrivateFetchSql)
	 */

	/* Construct cursor name from sequential value */
	sprintf(name, CURSOR_NAME_FORMAT, cursor_id++);

	/* Construct statement to declare cursor */
	initStringInfo(&cursor);
	appendStringInfo(&cursor, "DECLARE %s SCROLL CURSOR FOR %s", name, sql);
	fdw_private = lappend(fdw_private, makeString(cursor.data));

	/* Construct statement to fetch rows from cursor */
	initStringInfo(&cursor);
	appendStringInfo(&cursor, "FETCH %d FROM %s", fetch_count, name);
	fdw_private = lappend(fdw_private, makeString(cursor.data));

	/* Construct statement to reset cursor */
	initStringInfo(&cursor);
	appendStringInfo(&cursor, "MOVE ABSOLUTE 0 FROM %s", name);
	fdw_private = lappend(fdw_private, makeString(cursor.data));

	/* Construct statement to close cursor */
	initStringInfo(&cursor);
	appendStringInfo(&cursor, "CLOSE %s", name);
	fdw_private = lappend(fdw_private, makeString(cursor.data));

	/* Store FDW private information into FdwPlan */
	fdwplan->fdw_private = fdw_private;

	return fdwplan;
}

/*
 * pgsqlExplainForeignScan
 *		Produce extra output for EXPLAIN
 */
static void
pgsqlExplainForeignScan(ForeignScanState *node, ExplainState *es)
{
	FdwPlan	   *fdwplan;
	char	   *sql;

	fdwplan = ((ForeignScan *) node->ss.ps.plan)->fdwplan;
	sql = strVal(list_nth(fdwplan->fdw_private, FdwPrivateDeclareSql));
	ExplainPropertyText("Remote SQL", sql, es);
}

/*
 * pgsqlBeginForeignScan
 *		Initiate access to a foreign PostgreSQL table.
 */
static void
pgsqlBeginForeignScan(ForeignScanState *node, int eflags)
{
	PgsqlFdwExecutionState *festate;
	PGconn		   *conn;
	Oid				relid;
	ForeignTable   *table;
	ForeignServer  *server;
	UserMapping	   *user;
	TupleTableSlot *slot = node->ss.ss_ScanTupleSlot;

	/*
	 * Do nothing in EXPLAIN (no ANALYZE) case.  node->fdw_state stays NULL.
	 */
	if (eflags & EXEC_FLAG_EXPLAIN_ONLY)
		return;

	/*
	 * Save state in node->fdw_state.
	 */
	festate = (PgsqlFdwExecutionState *) palloc(sizeof(PgsqlFdwExecutionState));
	festate->fdwplan = ((ForeignScan *) node->ss.ps.plan)->fdwplan;

	/*
	 * Create context for per-scan tuplestore.
	 */
	festate->scan_cxt = AllocSetContextCreate(MessageContext,
											  "pgsql_fdw",
											  ALLOCSET_DEFAULT_MINSIZE,
											  ALLOCSET_DEFAULT_INITSIZE,
											  ALLOCSET_DEFAULT_MAXSIZE);

	/*
	 * Get connection to the foreign server.  Connection manager would
	 * establish new connection if necessary.
	 */
	relid = RelationGetRelid(node->ss.ss_currentRelation);
	table = GetForeignTable(relid);
	server = GetForeignServer(table->serverid);
	user = GetUserMapping(GetOuterUserId(), server->serverid);
	conn = GetConnection(server, user, true);
	festate->conn = conn;

	/* Result will be filled in first Iterate call. */
	festate->tuples = NULL;

	/* Allocate buffers for column values. */
	{
		TupleDesc	tupdesc = slot->tts_tupleDescriptor;
		festate->col_values = palloc(sizeof(char *) * tupdesc->natts);
		festate->attinmeta = TupleDescGetAttInMetadata(tupdesc);
	}

	/* Allocate buffers for query parameters. */
	{
		ParamListInfo	params = node->ss.ps.state->es_param_list_info;
		int				numParams = params ? params->numParams : 0;

		if (numParams > 0)
		{
			festate->param_types = palloc0(sizeof(Oid) * numParams);
			festate->param_values = palloc0(sizeof(char *) * numParams);
		}
		else
		{
			festate->param_types = NULL;
			festate->param_values = NULL;
		}
	}


	/* Store FDW-specific state into ForeignScanState */
	node->fdw_state = (void *) festate;

	return;
}

/*
 * pgsqlIterateForeignScan
 *		Retrieve next row from the result set, or clear tuple slot to indicate
 *		EOF.
 *
 *		Note that using per-query context when retrieving tuples from
 *		tuplestore to ensure that returned tuples can survive until next
 *		iteration because the tuple is released implicitly via ExecClearTuple.
 *		Retrieving a tuple from tuplestore in CurrentMemoryContext (it's a
 *		per-tuple context), ExecClearTuple will free dangling pointer.
 */
static TupleTableSlot *
pgsqlIterateForeignScan(ForeignScanState *node)
{
	PgsqlFdwExecutionState *festate;
	TupleTableSlot *slot = node->ss.ss_ScanTupleSlot;
	PGresult	   *res;
	MemoryContext	oldcontext = CurrentMemoryContext;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;


	/*
	 * If this is the first call after Begin, we need to execute remote query.
	 * If the query needs cursor, we declare a cursor at first call and fetch
	 * from it in later calls.
	 */
	if (festate->tuples == NULL)
		execute_query(node);

	/*
	 * If enough tuples are left in tuplestore, just return next tuple from it.
	 */
	MemoryContextSwitchTo(node->ss.ps.state->es_query_cxt);
	if (tuplestore_gettupleslot(festate->tuples, true, false, slot))
	{
		MemoryContextSwitchTo(oldcontext);
		return slot;
	}
	MemoryContextSwitchTo(oldcontext);

	/*
	 * Here we need to clear partial result and fetch next bunch of tuples from
	 * from the cursor for the scan.  If the fetch returns no tuple, the scan
	 * has reached the end.
	 */
	res = fetch_result(node);
	PG_TRY();
	{
		store_result(node, res);
		PQclear(res);
		res = NULL;
	}
	PG_CATCH();
	{
		PQclear(res);
		PG_RE_THROW();
	}
	PG_END_TRY();

	/*
	 * If we got more tuples from the server cursor, return next tuple from
	 * tuplestore.
	 */
	MemoryContextSwitchTo(node->ss.ps.state->es_query_cxt);
	if (tuplestore_gettupleslot(festate->tuples, true, false, slot))
	{
		MemoryContextSwitchTo(oldcontext);
		return slot;
	}
	MemoryContextSwitchTo(oldcontext);

	/* We don't have any result even in remote server cursor. */
	ExecClearTuple(slot);
	return slot;
}

/*
 * pgsqlReScanForeignScan
 *   - Restart this scan by resetting fetch location.
 */
static void
pgsqlReScanForeignScan(ForeignScanState *node)
{
	List	   *fdw_private;
	char	   *sql;
	PGconn	   *conn;
	PGresult   *res;
	PgsqlFdwExecutionState *festate;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;

	/* If we have not opened cursor yet, nothing to do. */
	if (festate->tuples == NULL)
		return;

	/* Discard fetch results if any. */
	tuplestore_clear(festate->tuples);

	/* Reset cursor */
	fdw_private = festate->fdwplan->fdw_private;
	conn = festate->conn;
	sql = strVal(list_nth(fdw_private, FdwPrivateResetSql));
	res = PQexec(conn, sql);
	PG_TRY();
	{
		if (PQresultStatus(res) != PGRES_COMMAND_OK)
		{
			ereport(ERROR,
					(errmsg("could not rewind cursor"),
					 errdetail("%s", PQerrorMessage(conn)),
					 errhint("%s", sql)));
		}
		PQclear(res);
		res = NULL;
	}
	PG_CATCH();
	{
		PQclear(res);
		PG_RE_THROW();
	}
	PG_END_TRY();
}

/*
 * pgsqlEndForeignScan
 *		Finish scanning foreign table and dispose objects used for this scan
 */
static void
pgsqlEndForeignScan(ForeignScanState *node)
{
	List	   *fdw_private;
	char	   *sql;
	PGconn	   *conn;
	PGresult   *res;
	PgsqlFdwExecutionState *festate;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;

	/* if festate is NULL, we are in EXPLAIN; nothing to do */
	if (festate == NULL)
		return;

	/* If we have not opened cursor yet, nothing to do. */
	if (festate->tuples == NULL)
		return;

	/* Discard fetch results */
	tuplestore_end(festate->tuples);
	festate->tuples = NULL;

	/* Close cursor */
	fdw_private = festate->fdwplan->fdw_private;
	conn = festate->conn;
	sql = strVal(list_nth(fdw_private, FdwPrivateCloseSql));
	res = PQexec(conn, sql);
	PG_TRY();
	{
		if (PQresultStatus(res) != PGRES_COMMAND_OK)
		{
			ereport(ERROR,
					(errmsg("could not close cursor"),
					 errdetail("%s", PQerrorMessage(conn)),
					 errhint("%s", sql)));
		}
		PQclear(res);
		res = NULL;
	}
	PG_CATCH();
	{
		PQclear(res);
		PG_RE_THROW();
	}
	PG_END_TRY();

	ReleaseConnection(festate->conn);

	MemoryContextDelete(festate->scan_cxt);
	festate->scan_cxt = NULL;
}

/*
 * pgsqlAnalyzeForeignTable
 *		Update statistics of given foreign table.
 */
static void
pgsqlAnalyzeForeignTable(Relation relation,
						 VacuumStmt *vacstmt,
						 int elevel)
{
	int				nstats = 0;
	MemoryContext	anl_context;
	MemoryContext	caller_context;
	const char	   *nspname;
	const char	   *relname;
	ForeignTable   *table;
	ForeignServer  *server;
	UserMapping	   *user;
	PGconn		   *conn;
	PGresult	   *res;
	StringInfoData	sql;
	char		   *endp;
	double			reltuples;
	int				relpages;
	int				attr_cnt;
	int				tcnt;
	int				i;
	VacAttrStats  **vacattrstats;

	/* Use short-lived expression context. */
	anl_context = AllocSetContextCreate(CurrentMemoryContext,
										"Analyze",
										ALLOCSET_DEFAULT_MINSIZE,
										ALLOCSET_DEFAULT_INITSIZE,
										ALLOCSET_DEFAULT_MAXSIZE);
	caller_context = MemoryContextSwitchTo(anl_context);

	/*
	 * Establish connection against the foreign server to retrieve remote
	 * statistics data.
	 */
	table = GetForeignTable(relation->rd_id);
	server = GetForeignServer(table->serverid);
	user = GetUserMapping(GetOuterUserId(), table->serverid);
	conn = GetConnection(server, user, false);

	/* Construct SQL statement to retrieve per-relation statistics. */
	nspname = GetFdwOptionValue(InvalidOid, InvalidOid, relation->rd_id,
								InvalidAttrNumber, "nspname");
	if (nspname == NULL)
		nspname = get_namespace_name(get_rel_namespace(relation->rd_id));
	nspname = quote_identifier(nspname);
	relname = GetFdwOptionValue(InvalidOid, InvalidOid, relation->rd_id,
								InvalidAttrNumber, "relname");
	if (relname == NULL)
		relname = get_rel_name(relation->rd_id);
	relname = quote_identifier(relname);

	/*
	 * Get per-relation statistics from remote pg_class.  We don't fetch
	 * relallvisible, which is used for estimation about index-only scan,
	 * because index-only scan is not available on foreign tables.
	 * 
	 * XXX Should we get summary of whole inherit-tree for inherited table?
	 */
	initStringInfo(&sql);
	appendStringInfo(&sql,
					 "SELECT relpages, "
					 "       reltuples "
					 "  FROM pg_catalog.pg_class "
					 " WHERE oid = '%s.%s'::regclass",
					 nspname, relname);

	res = PQexec(conn, sql.data);

	if (PQresultStatus(res) != PGRES_TUPLES_OK)
		ereport(ERROR,
				(errmsg("could not get per-table statistics"),
				 errdetail("%s", PQerrorMessage(conn))));
	if (PQntuples(res) != 1)
		ereport(ERROR,
				(errmsg("invalid number of tuples: %d", PQntuples(res))));

	relpages = strtol(PGRES_VAL0(0), &endp, 10);
	if (*endp != '\0')
		ereport(ERROR,
				(errmsg("invalid format for relpages: %s", PGRES_VAL0(0))));
	reltuples = strtod(PGRES_VAL0(1), &endp);
	if (*endp != '\0')
		ereport(ERROR,
				(errmsg("invalid format for reltuples: %s", PGRES_VAL0(1))));
	PQclear(res);

	/*
	 * Get per-column statistics from remote pg_statistic.  Note that some of
	 * local columns might be dropped.
	 */
	if (vacstmt->va_cols != NIL)
	{
		ListCell   *le;

		vacattrstats = (VacAttrStats **) palloc(list_length(vacstmt->va_cols) *
												sizeof(VacAttrStats *));
		tcnt = 0;
		foreach(le, vacstmt->va_cols)
		{
			char	   *col = strVal(lfirst(le));

			i = attnameAttNum(relation, col, false);
			if (i == InvalidAttrNumber)
				ereport(ERROR,
						(errcode(ERRCODE_UNDEFINED_COLUMN),
					errmsg("column \"%s\" of relation \"%s\" does not exist",
						   col, RelationGetRelationName(relation))));
			vacattrstats[tcnt] = examine_attribute(relation, i, NULL);
			if (vacattrstats[tcnt] != NULL)
				tcnt++;
		}
		attr_cnt = tcnt;
	}
	else
	{
		attr_cnt = relation->rd_att->natts;
		vacattrstats = (VacAttrStats **)
			palloc(attr_cnt * sizeof(VacAttrStats *));
		tcnt = 0;
		for (i = 1; i <= attr_cnt; i++)
		{
			vacattrstats[tcnt] = examine_attribute(relation, i, NULL);
			if (vacattrstats[tcnt] != NULL)
				tcnt++;
		}
		attr_cnt = tcnt;
	}

	/* Retrieve statistics for each requested column from foreign server. */
	for (i = 0; i < attr_cnt; i++)
	{
		const char	   *colname;
		StringInfoData	sql_stat;

		colname = GetFdwOptionValue(InvalidOid, InvalidOid, relation->rd_id,
									vacattrstats[i]->attr->attnum, "colname");
		if (colname == NULL)
			colname = get_attname(relation->rd_id,
								  vacattrstats[i]->attr->attnum);
		/* No need to quote column name here. */
		initStringInfo(&sql_stat);
		appendStringInfo(&sql_stat,
						 "SELECT starelid,"
						 "       staattnum,"
						 /* fixed value is used for stainherit, see below */
						 "       stanullfrac,"
						 "       stawidth,"
						 "       stadistinct,"
						 "       stakind1,"
						 "       stakind2,"
						 "       stakind3,"
						 "       stakind4,"
						 "       staop1,"
						 "       staop2,"
						 "       staop3,"
						 "       staop4,"
						 "       stanumbers1,"
						 "       stanumbers2,"
						 "       stanumbers3,"
						 "       stanumbers4,"
						 "       stavalues1,"
						 "       stavalues2,"
						 "       stavalues3,"
						 "       stavalues4"
						 "  FROM pg_catalog.pg_statistic "
						 " WHERE starelid = '%s.%s'::regclass "
						 "   AND staattnum = ("
						 "           SELECT attnum "
						 "             FROM pg_catalog.pg_attribute "
						 "            WHERE attrelid = '%s.%s'::regclass "
						 "              AND attname = '%s')",
						 nspname, relname, nspname, relname, colname);

		/*
		 * If foreign server is PG 9.0 or later, inherited tables have two sets
		 * of statistics; one contains information about the table itself, and
		 * another contains information about whole inherit-tree topped by the
		 * table.
		 *
		 * We prefer statistics of whole inherit-tree if any, because parent
		 * tables return contents of descendants too.  However, we use "false"
		 * for stainherit of local statistics of foreign tables, because
		 * foreign tables can't be inherited.
		 *
		 * With additional condition below, only the row which has false in
		 * stainherit would be returned if any.
		 */
		if (PQserverVersion(conn) >= 90000)
			appendStringInfo(&sql_stat, "ORDER BY stainherit DESC LIMIT 1");

		res = PQexec(conn, sql_stat.data);
		pfree(sql_stat.data);
		if (PQresultStatus(res) != PGRES_TUPLES_OK)
			ereport(ERROR,
					(errmsg("could not get per-column statistics"),
					 errdetail("%s", PQerrorMessage(conn))));

		/* If remote table doesn't have any statistic, skip this column. */
		if (PQntuples(res) > 0)
		{
			store_remote_stats(res, vacattrstats[i]);
			nstats++;
		}

		PQclear(res);
	}

	/* Disconnect from foreign server. */
	ReleaseConnection(conn);

	/* Update local pg_statistic with retrieved statistics. */
	vac_update_attstats(RelationGetRelid(relation),
						false,
						attr_cnt,
						vacattrstats);

	/*
	 * Update local pg_class with retrieved statistics.  We set relallvisible
	 * to 0 always.
	 */
	vac_update_relstats(relation,
						relpages,
						reltuples,
						0,
						false,
						InvalidTransactionId);

	/* Restore memory context. */
	MemoryContextSwitchTo(caller_context);
	MemoryContextDelete(anl_context);

	/* Print verbose messages. */
	if (vacstmt->options & VACOPT_VERBOSE)
		elog(INFO, "%s: relpages=%d, reltuples=%f, %d per-column stats copied",
			 vacstmt->relation->relname,
			 relpages,
			 reltuples,
			 nstats);
}

/*
 * Read text representation of remote statistics from res and store them into
 * stats with conversion to internal representation.
 */
static void
store_remote_stats(PGresult *res, VacAttrStats *stats)
{
	Type			float4array_type;
	Type			textarray_type;
	Oid				elemtypeid;
	Type			elemtype;
	Oid				in_func_oid;
	Form_pg_type	attrtype;
	Oid				typeioparam;
	FmgrInfo		in_function;

	int				i_stanullfrac;
	int				i_stawidth;
	int				i_stadistinct;
	int				i_stakind[STATISTIC_NUM_SLOTS];
	int				i_staop[STATISTIC_NUM_SLOTS];
	int				i_stanumbers[STATISTIC_NUM_SLOTS];
	int				i_stavalues[STATISTIC_NUM_SLOTS];
	int				i;

	/* Get information of actual type of array element. */
	elemtypeid = stats->attr->atttypid;
	elemtype = typeidType(elemtypeid);
	getTypeInputInfo(elemtypeid, &in_func_oid, &typeioparam);
	fmgr_info(in_func_oid, &in_function);

	attrtype = (Form_pg_type) palloc(sizeof(Form_pg_type));
	memcpy(attrtype, GETSTRUCT(elemtype), sizeof(FormData_pg_type));

	i_stanullfrac = PQfnumber(res, "stanullfrac");
	i_stawidth = PQfnumber(res, "stawidth");
	i_stadistinct = PQfnumber(res, "stadistinct");
	i_stakind[0] = PQfnumber(res, "stakind1");
	i_stakind[1] = PQfnumber(res, "stakind2");
	i_stakind[2] = PQfnumber(res, "stakind3");
	i_stakind[3] = PQfnumber(res, "stakind4");
	i_staop[0] = PQfnumber(res, "staop1");
	i_staop[1] = PQfnumber(res, "staop2");
	i_staop[2] = PQfnumber(res, "staop3");
	i_staop[3] = PQfnumber(res, "staop4");
	i_stanumbers[0] = PQfnumber(res, "stanumbers1");
	i_stanumbers[1] = PQfnumber(res, "stanumbers2");
	i_stanumbers[2] = PQfnumber(res, "stanumbers3");
	i_stanumbers[3] = PQfnumber(res, "stanumbers4");
	i_stavalues[0] = PQfnumber(res, "stavalues1");
	i_stavalues[1] = PQfnumber(res, "stavalues2");
	i_stavalues[2] = PQfnumber(res, "stavalues3");
	i_stavalues[3] = PQfnumber(res, "stavalues4");

	/*
	 * If statistics are found, mark this column to be updated later
	 * and fill with retrieved remote statistics.
	 *
	 * For scalar values, just copying values is enough, but we have to
	 * be conscious the types of elements for array values.  libpq
	 * gives us an array value in a string representation, such as
	 * "{foo, bar}", so we need to (1)separate an array string into
	 * elements, (2)convert each element into Datum representation, and
	 * (3) create a Datum representation of array from Datum elements.
	 *
	 * Note: OID of operator (staopN) MUST match between remote and local.
	 */
	stats->stats_valid = true;
	stats->stanullfrac = strtof(PGRES_VAL0(i_stanullfrac), NULL);
	stats->stawidth = strtol(PGRES_VAL0(i_stawidth), NULL, 10);
	stats->stadistinct = strtof(PGRES_VAL0(i_stadistinct), NULL);

	/* Get type information for stanumbers and stavalues. */
	float4array_type = typeidType(FLOAT4ARRAYOID);
	textarray_type = typeidType(TEXTARRAYOID);

	for (i = 0; i < STATISTIC_NUM_SLOTS; i++)
	{
		int			atttypmod = 0;
		int			j;
		Datum		tnumbers = 0;
		ArrayType  *array;
		int			nitems;
		bits8	   *bitmap;
		int			bitmask;
		char	   *p;

		stats->stakind[i] = strtol(PGRES_VAL0(i_stakind[i]), NULL, 10);
		stats->staop[i] = strtol(PGRES_VAL0(i_staop[i]), NULL, 10);

		/*
		 * Extrace element values of stanumbersN from text representation of
		 * array, if any.
		 */
		if (!PGRES_NULL0(i_stanumbers[i]))
		{
			/* Separate array string into each float4 elements. */
			tnumbers = stringTypeDatum(float4array_type,
									   PGRES_VAL0(i_stanumbers[i]),
									   -1);
			array = DatumGetArrayTypeP(tnumbers);
			nitems = ArrayGetNItems(ARR_NDIM(array), ARR_DIMS(array));

			bitmap = ARR_NULLBITMAP(array);
			bitmask = 1;

			stats->numnumbers[i] = nitems;
			stats->stanumbers[i] = (float4 *) palloc(nitems * sizeof(float4));
			for (j = 0; j < nitems; j++)
			{
				p = ARR_DATA_PTR(array);

				if (bitmap && (*bitmap & bitmask) == 0)
				{
					/* nulls can be ignored */
				}
				else
				{
					stats->stanumbers[i][j] = * (float4 *) ARR_DATA_PTR(array);
					p = p + sizeof(float4);
					p = (char *) att_align_nominal(p, 'i');
				}

				/* advance bitmap pointer if any */
				if (bitmap)
				{
					bitmask <<= 1;
					if (bitmask == 0x100)
					{
						bitmap++;
						bitmask = 1;
					}
				}
			}
		}
		else
		{
			stats->numnumbers[i] = 0;
			stats->stanumbers[i] = NULL;
		}

		/*
		 * Extrace element values of stavaluesN from text representation of
		 * array, if any.
		 */
		if (!PGRES_NULL0(i_stavalues[i]))
		{
			/*
			 * Separate array string into each elements string, and convert
			 * each element to actual column data type.
			 */
			tnumbers = stringTypeDatum(textarray_type,
									   PGRES_VAL0(i_stavalues[i]),
									   -1);
			array = DatumGetArrayTypeP(tnumbers);
			nitems = ArrayGetNItems(ARR_NDIM(array), ARR_DIMS(array));

			bitmap = ARR_NULLBITMAP(array);
			bitmask = 1;
			p = ARR_DATA_PTR(array);

			stats->numvalues[i] = nitems;
			stats->stavalues[i] = (Datum *) palloc(nitems * sizeof(Datum));

			for (j = 0; j < nitems; j++)
			{
				char	   *string;

				if (bitmap && (*bitmap & bitmask) == 0)
				{
					/* nulls can be ignored */
				}
				else
				{
					string = text_to_cstring((text *) p);
					stats->stavalues[i][j] = InputFunctionCall(&in_function,
															   string,
															   typeioparam,
															   atttypmod);
					p = p + VARSIZE_ANY(p);
					p = (char *) att_align_nominal(p, 'i');
				}

				/* advance bitmap pointer if any */
				if (bitmap)
				{
					bitmask <<= 1;
					if (bitmask == 0x100)
					{
						bitmap++;
						bitmask = 1;
					}
				}
			}
		}
		else
		{
			stats->numvalues[i] = 0;
			stats->stavalues[i] = NULL;
		}
	}

	/* Clean up */
	ReleaseSysCache(elemtype);
	ReleaseSysCache(textarray_type);
	ReleaseSysCache(float4array_type);
}

/*
 * Estimate costs of scanning a foreign table.
 */
static void
estimate_costs(PlannerInfo *root, RelOptInfo *baserel,
			   const char *sql, Oid serverid,
			   Cost *startup_cost, Cost *total_cost)
{
	ListCell	   *lc;
	ForeignServer  *server;
	UserMapping	   *user;
	PGconn		   *conn = NULL;
	PGresult	   *res = NULL;
	StringInfoData  buf;
	char		   *plan;
	char		   *p;
	int				n;
	double			nrows;

	/*
	 * If the baserestrictinfo contains any Param node with paramkind
	 * PARAM_EXTERNAL, we need result of EXPLAIN for EXECUTE statement, not for
	 * simple SELECT statement.  However, we can't get actual parameter values
	 * here, so we use fixed and large costs as second best so that planner
	 * tend to choose custom plan.
	 *
	 * See comments in plancache.c for details of custom plan.
	 */
	foreach(lc, baserel->baserestrictinfo)
	{
		RestrictInfo	   *rs = (RestrictInfo *) lfirst(lc);
		if (contain_ext_param((Node *) rs->clause))
		{
			*startup_cost = CONNECTION_COSTS;
			*total_cost = 10000;	/* groundless large costs */

			return;
		}
	}

	/*
	 * Get connection to the foreign server.  Connection manager would
	 * establish new connection if necessary.
	 */
	server = GetForeignServer(serverid);
	user = GetUserMapping(GetOuterUserId(), server->serverid);
	conn = GetConnection(server, user, false);
	initStringInfo(&buf);
	appendStringInfo(&buf, "EXPLAIN %s", sql);

	PG_TRY();
	{
		res = PQexec(conn, buf.data);
		if (PQresultStatus(res) != PGRES_TUPLES_OK || PQntuples(res) == 0)
		{
			char *msg;

			msg = pstrdup(PQerrorMessage(conn));
			ereport(ERROR,
					(errmsg("could not execute EXPLAIN for cost estimation"),
					 errdetail("%s", msg),
					 errhint("%s", sql)));
		}

		/*
		 * Find estimation portion from top plan node. Here we search opening
		 * parentheses from the end of the line to avoid finding unexpected
		 * parentheses.
		 */
		plan = PQgetvalue(res, 0, 0);
		p = strrchr(plan, '(');
		if (p == NULL)
			elog(ERROR, "wrong EXPLAIN output: %s", plan);
		n = sscanf(p,
				   "(cost=%lf..%lf rows=%lf width=%d)",
				   startup_cost, total_cost, &nrows, &baserel->width);
		if (n != 4)
			elog(ERROR, "could not get estimation from EXPLAIN output");

		PQclear(res);
		res = NULL;
		ReleaseConnection(conn);
	}
	PG_CATCH();
	{
		PQclear(res);
		PG_RE_THROW();
	}
	PG_END_TRY();

	/*
	 * Selectivity of quals which have not been pushed down should be also
	 * considered.
	 */
	baserel->rows = nrows * clauselist_selectivity(root,
												   baserel->baserestrictinfo,
												   0,
												   JOIN_INNER,
												   NULL);

	/* add cost to establish connection. */
	*startup_cost += CONNECTION_COSTS;
	*total_cost += CONNECTION_COSTS;

	/* add cost to transfer result. */
	*total_cost += TRANSFER_COSTS_PER_BYTE * baserel->width * baserel->tuples;
	*total_cost += cpu_tuple_cost * baserel->tuples;
}

/*
 * Execute remote query with current parameters.
 */
static void
execute_query(ForeignScanState *node)
{
	FdwPlan		   *fdwplan;
	PgsqlFdwExecutionState *festate;
	ParamListInfo	params = node->ss.ps.state->es_param_list_info;
	int				numParams = params ? params->numParams : 0;
	Oid			   *types = NULL;
	const char	  **values = NULL;
	char		   *sql;
	PGconn		   *conn;
	PGresult	   *res;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;
	types = festate->param_types;
	values = festate->param_values;

	/*
	 * Construct parameter array in text format.  We don't release memory for
	 * the arrays explicitly, because the memory usage would not be very large,
	 * and anyway they will be released in context cleanup.
	 */
	if (numParams > 0)
	{
		int i;

		for (i = 0; i < numParams; i++)
		{
			types[i] = params->params[i].ptype;
			if (params->params[i].isnull)
				values[i] = NULL;
			else
			{
				Oid			out_func_oid;	
				bool		isvarlena;
				FmgrInfo	func;

				getTypeOutputInfo(types[i], &out_func_oid, &isvarlena);
				fmgr_info(out_func_oid, &func);
				values[i] = OutputFunctionCall(&func, params->params[i].value);
			}
		}
	}

	/*
	 * Execute remote query with parameters.
	 */
	conn = festate->conn;
	fdwplan = ((ForeignScan *) node->ss.ps.plan)->fdwplan;
	sql = strVal(list_nth(fdwplan->fdw_private, FdwPrivateDeclareSql));
	res = PQexecParams(conn, sql, numParams, types, values, NULL, NULL, 0);
	PG_TRY();
	{
		/*
		 * If the query has failed, reporting details is enough here.
		 * Connection(s) which are used by this query (at least used by
		 * pgsql_fdw) will be cleaned up by the foreign connection manager.
		 */
		if (PQresultStatus(res) != PGRES_COMMAND_OK)
		{
			ereport(ERROR,
					(errmsg("could not declare cursor"),
					 errdetail("%s", PQerrorMessage(conn)),
					 errhint("%s", sql)));
		}

		/* Discard result of CURSOR statement and fetch first bunch. */
		PQclear(res);
		res = fetch_result(node);

		/*
		 * Store the result of the query into tuplestore.
		 * We must release PGresult here to avoid memory leak.
		 */
		store_result(node, res);
		PQclear(res);
		res = NULL;
	}
	PG_CATCH();
	{
		PQclear(res);
		PG_RE_THROW();
	}
	PG_END_TRY();
}

/*
 * Fetch next partial result from remote server.
 */
static PGresult *
fetch_result(ForeignScanState *node)
{
	PgsqlFdwExecutionState *festate;
	List		   *fdw_private;
	char		   *sql;
	PGconn		   *conn;
	PGresult	   *res;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;

	/* retrieve information for fetching result. */
	fdw_private = festate->fdwplan->fdw_private;
	sql = strVal(list_nth(fdw_private, FdwPrivateFetchSql));
	conn = festate->conn;
	res = PQexec(conn, sql);
	if (PQresultStatus(res) != PGRES_TUPLES_OK)
	{
		ereport(ERROR,
				(errmsg("could not fetch rows from foreign server"),
				 errdetail("%s", PQerrorMessage(conn)),
				 errhint("%s", sql)));
	}

	return res;
}

/*
 * Create tuples from PGresult and store them into tuplestore.
 */
static void
store_result(ForeignScanState *node, PGresult *res)
{
	int					rows;
	int					row;
	int					i;
	int					nfields;
	int					attnum;		/* number of non-dropped columns */
	Form_pg_attribute  *attrs;
	TupleTableSlot *slot = node->ss.ss_ScanTupleSlot;
	TupleDesc	tupdesc = slot->tts_tupleDescriptor;
	PgsqlFdwExecutionState *festate;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;
	rows = PQntuples(res);
	nfields = PQnfields(res);
	attrs = tupdesc->attrs;

	/* First, ensure that the tuplestore is empty. */
	if (festate->tuples == NULL)
	{
		MemoryContext	oldcontext = CurrentMemoryContext;

		/*
		 * Create tuplestore to store result of the query in per-query context.
		 * Note that we use this memory context to avoid memory leak in error
		 * cases.
		 */
		MemoryContextSwitchTo(festate->scan_cxt);
		festate->tuples = tuplestore_begin_heap(false, false, work_mem);
		MemoryContextSwitchTo(oldcontext);
	}
	else
	{
		/* We already have tuplestore, just need to clear contents of it. */
		tuplestore_clear(festate->tuples);
	}
	

	/* count non-dropped columns */
	for (attnum = 0, i = 0; i < tupdesc->natts; i++)
		if (!attrs[i]->attisdropped)
			attnum++;

	/* check result and tuple descriptor have the same number of columns */
	if (attnum > 0 && attnum != nfields)
		ereport(ERROR,
				(errcode(ERRCODE_DATATYPE_MISMATCH),
				 errmsg("remote query result rowtype does not match "
						"the specified FROM clause rowtype"),
				 errdetail("expected %d, actual %d", attnum, nfields)));

	/* put a tuples into the slot */
	for (row = 0; row < rows; row++)
	{
		int			j;
		HeapTuple	tuple;

		for (i = 0, j = 0; i < tupdesc->natts; i++)
		{
			/* skip dropped columns. */
			if (attrs[i]->attisdropped)
			{
				festate->col_values[i] = NULL;
				continue;
			}

			if (PQgetisnull(res, row, j))
				festate->col_values[i] = NULL;
			else
				festate->col_values[i] = PQgetvalue(res, row, j);
			j++;
		}

		/*
		 * Build the tuple and put it into the slot.
		 * We don't have to free the tuple explicitly because it's been
		 * allocated in the per-tuple context.
		 */
		tuple = BuildTupleFromCStrings(festate->attinmeta, festate->col_values);
		tuplestore_puttuple(festate->tuples, tuple);
	}

	tuplestore_donestoring(festate->tuples);
}

/*
 * contain_ext_param
 *	  Recursively search for Param nodes within a clause.
 *
 *	  Returns true if any parameter reference node with relkind PARAM_EXTERN
 *	  found.
 *
 * This does not descend into subqueries, and so should be used only after
 * reduction of sublinks to subplans, or in contexts where it's known there
 * are no subqueries.  There mustn't be outer-aggregate references either.
 *
 * XXX: These functions could be in core, src/backend/optimizer/util/clauses.c.
 */
static bool
contain_ext_param(Node *clause)
{
	return contain_ext_param_walker(clause, NULL);
}

static bool
contain_ext_param_walker(Node *node, void *context)
{
	if (node == NULL)
		return false;
	if (IsA(node, Param))
	{
		Param	   *param = (Param *) node;

		if (param->paramkind == PARAM_EXTERN)
			return true;		/* abort the tree traversal and return true */
	}
	return expression_tree_walker(node, contain_ext_param_walker, context);
}
